LIBRARY     SWMM5.DLL

EXPORTS
    swmm_close                    = _swmm_close@0
    swmm_end                      = _swmm_end@0
    swmm_getError                 = _swmm_getError@8
    swmm_getMassBalErr            = _swmm_getMassBalErr@12
    swmm_getVersion               = _swmm_getVersion@0
    swmm_getWarnings              = _swmm_getWarnings@0
    swmm_open                     = _swmm_open@12
    swmm_report                   = _swmm_report@0
    swmm_run                      = _swmm_run@12
    swmm_start                    = _swmm_start@4
    swmm_step                     = _swmm_step@4
