//-----------------------------------------------------------------------------
//   headers.h
//
//   Project: EPA SWMM5
//   Version: 5.1
//   Date:    03/20/14   (Build 5.1.001)
//   Author:  L. Rossman
//
//   Header files included in most SWMM5 modules.
//
//   DO NOT CHANGE THE ORDER OF THE #INCLUDE STATEMENTS
//-----------------------------------------------------------------------------

#include <stdio.h>
#include "consts.h"
#include "macros.h"
#include "enums.h"
#include "error.h"
#include "datetime.h"
#include "objects.h"
#include "funcs.h"
#include "text.h"
#include "keywords.h"
#define  EXTERN extern
#include "globals.h"
#include "lid.h"